package controllers

import (
	"bytes"
	"fmt"
	"text/template"

	webservicescernchv1alpha1 "gitlab.cern.ch/paas-tools/operators/static-site-operator/api/v1alpha1"
	v1 "k8s.io/api/core/v1"
)

const httpdConfigTemplate string = `<VirtualHost *:8080>
  ServerName {{ .Host }}

  ProxyAddHeaders Off
  ProxyPass / {{ .GitlabPagesURL }}
  ProxyPassReverse / {{ .GitlabPagesURL }}
  ProxyPreserveHost On

  {{ if (not .Anonymous) -}}
  OIDCProviderMetadataURL {{ printf "%s/.well-known/openid-configuration" .OIDCSecret.IssuerURL }}
  OIDCClientID {{ .OIDCSecret.ClientID }}
  OIDCClientSecret {{ .OIDCSecret.ClientSecret }}

  # OIDCRedirectURI is a vanity URL that must point to a path protected by this module but must NOT point to any content
  OIDCRedirectURI {{ printf "https://%s/redirect-uri" .Host }}
  OIDCCryptoPassphrase {{ .CryptoPassPhrase }}

  <Location / >
    AuthType openid-connect
    Require valid-user
  </Location>
  {{- end }}
</VirtualHost>`

var templ *template.Template

func init() {
	templ = template.Must(template.New("httpd-config.template").Parse(httpdConfigTemplate))
}

func generateSecretKey(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite) string {
	return fmt.Sprintf("%s-static-site.conf", gitlabPagesSite.ObjectMeta.Name)
}

type templateData struct {
	webservicescernchv1alpha1.GitlabPagesSiteSpec

	GitlabPagesURL   string
	CryptoPassPhrase string
	OIDCSecret       struct {
		ClientID, ClientSecret, IssuerURL string
	}
}

func generateHttpdVhostConfiguration(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite, oidcSecret *v1.Secret, gitlabPagesURL string, cryptoPassPhrase string) ([]byte, error) {
	var buf bytes.Buffer

	data := templateData{
		GitlabPagesSiteSpec: gitlabPagesSite.Spec,
		GitlabPagesURL:      gitlabPagesURL,
		CryptoPassPhrase:    cryptoPassPhrase,
	}

	data.OIDCSecret.ClientID = string(oidcSecret.Data["clientID"])
	data.OIDCSecret.ClientSecret = string(oidcSecret.Data["clientSecret"])
	data.OIDCSecret.IssuerURL = string(oidcSecret.Data["issuerURL"])

	if err := templ.Execute(&buf, data); err != nil {
		return []byte(""), err
	}

	return []byte(buf.String()), nil
}
