package controllers

import (
	"context"
	"fmt"

	routev1 "github.com/openshift/api/route/v1"
	authzalpha1 "gitlab.cern.ch/paas-tools/operators/authz-operator/api/v1alpha1"
	webservicescernchv1alpha1 "gitlab.cern.ch/paas-tools/operators/static-site-operator/api/v1alpha1"
	corev1 "k8s.io/api/core/v1"
	kerrors "k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/apimachinery/pkg/util/intstr"
	"sigs.k8s.io/controller-runtime/pkg/client"
)

func (r *GitlabPagesSiteReconciler) ensureHttpdSecret(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite) (*corev1.Secret, error) {
	secret := &corev1.Secret{}
	err := r.Get(
		context.TODO(),
		types.NamespacedName{Namespace: r.Namespace, Name: r.VirtualHostsSecretName},
		secret,
	)

	if err != nil {
		return nil, err
	}

	return secret, nil
}

func (r *GitlabPagesSiteReconciler) getApplicationRegistration(ctx context.Context, gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite) (*authzalpha1.ApplicationRegistration, error) {
	appReg := &authzalpha1.ApplicationRegistration{}
	appRegistrationName := types.NamespacedName{Namespace: gitlabPagesSite.Namespace, Name: gitlabPagesSite.Name}

	if err := r.Get(ctx, appRegistrationName, appReg); err != nil {
		if kerrors.IsNotFound(err) {
			return nil, fmt.Errorf("ApplicationRegistration %v was not found, either it wasn't created or it was deleted", appRegistrationName.Name)
		} else {
			r.logger.Error(err, "Failed to fetch ApplicationRegistration")
			return nil, err
		}
	}
	return appReg, nil
}

func (r *GitlabPagesSiteReconciler) ensureGitlabPagesSiteHttpdConfig(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite, secret *corev1.Secret, oidcSecret *corev1.Secret) error {
	if secret.Data == nil {
		secret.Data = map[string][]byte{}
	}

	httpdConfig, err := generateHttpdVhostConfiguration(gitlabPagesSite, oidcSecret, r.GitlabPagesURL, r.CryptoPassPhrase)
	if err != nil {
		return err
	}

	secret.Data[generateSecretKey(gitlabPagesSite)] = httpdConfig

	err = r.Update(context.TODO(), secret, &client.UpdateOptions{})
	if err != nil {
		return err
	}

	return nil
}

// ensureRoute ensures the existence of a Route for a given GitlabPagesSite
func (r *GitlabPagesSiteReconciler) ensureRoute(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite) error {
	_, err := r.getRoute(gitlabPagesSite)

	if err != nil {
		if kerrors.IsNotFound(err) {
			route, err := r.newRouteForGitlabPagesSite(gitlabPagesSite)
			if err != nil {
				return err
			}

			if err := r.createRoute(route); err != nil {
				return err
			}
		}
	}

	return nil
}

func (r *GitlabPagesSiteReconciler) newRouteForGitlabPagesSite(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite) (*routev1.Route, error) {
	authenticatingProxyService := &corev1.Service{}

	err := r.Get(context.TODO(), types.NamespacedName{Namespace: r.Namespace, Name: r.AuthProxyServiceName}, authenticatingProxyService)
	if err != nil {
		return nil, err
	}

	route := &routev1.Route{
		ObjectMeta: metav1.ObjectMeta{
			Name:      gitlabPagesSite.Name,
			Namespace: r.Namespace,
			Annotations: map[string]string{
				"webservices.cern.ch/static-site-owner": types.NamespacedName{Name: gitlabPagesSite.Name, Namespace: gitlabPagesSite.Namespace}.String(),
			},
		},
		Spec: routev1.RouteSpec{
			TLS: &routev1.TLSConfig{
				InsecureEdgeTerminationPolicy: "Redirect",
				Termination:                   "edge",
			},
			To: routev1.RouteTargetReference{
				Kind: "Service",
				Name: authenticatingProxyService.ObjectMeta.Name,
			},
			Host: gitlabPagesSite.Spec.Host,
			Port: &routev1.RoutePort{
				TargetPort: intstr.FromInt(int(authenticatingProxyService.Spec.Ports[0].Port)),
			},
		},
	}

	return route, nil
}

func (r *GitlabPagesSiteReconciler) getRoute(gitlabPagesSite *webservicescernchv1alpha1.GitlabPagesSite) (route *routev1.Route, err error) {
	route = &routev1.Route{}
	key := types.NamespacedName{
		Name:      gitlabPagesSite.Name,
		Namespace: r.Namespace,
	}

	if err = r.Get(context.TODO(), key, route); err != nil {
		if kerrors.IsNotFound(err) {
			return nil, err
		}
		return nil, err
	}

	return route, nil
}

// createRoute creates a route for the given site
func (r *GitlabPagesSiteReconciler) createRoute(route *routev1.Route) error {
	if err := r.Create(context.TODO(), route); err != nil {
		return fmt.Errorf("failed to create route: %v", err)
	}
	return nil
}
