/*
Copyright 2021.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controllers

import (
	"context"
	"fmt"
	"reflect"
	"strings"
	"time"

	"github.com/go-logr/logr"
	routev1 "github.com/openshift/api/route/v1"
	v1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/types"

	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
	"sigs.k8s.io/controller-runtime/pkg/handler"
	"sigs.k8s.io/controller-runtime/pkg/log"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"
	"sigs.k8s.io/controller-runtime/pkg/source"

	authzalpha1 "gitlab.cern.ch/paas-tools/operators/authz-operator/api/v1alpha1"
	webservicescernchv1alpha1 "gitlab.cern.ch/paas-tools/operators/static-site-operator/api/v1alpha1"
)

const gitlabPagesSiteFinalizer string = "webservices.cern.ch/static-site-finalizer"

// GitlabPagesSiteReconciler reconciles a GitlabPagesSite object
type GitlabPagesSiteReconciler struct {
	client.Client
	Scheme                 *runtime.Scheme
	Namespace              string
	VirtualHostsSecretName string
	GitlabPagesURL         string
	AuthProxyServiceName   string
	CryptoPassPhrase       string
	logger                 logr.Logger
}

//+kubebuilder:rbac:groups=webservices.cern.ch,resources=gitlabpagessites,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=webservices.cern.ch,resources=gitlabpagessites/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=webservices.cern.ch,resources=gitlabpagessites/finalizers,verbs=update
//+kubebuilder:rbac:groups=webservices.cern.ch,resources=applicationregistrations,verbs=get;list;watch
//+kubebuilder:rbac:groups=route.openshift.io,resources=routes,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=route.openshift.io,resources=routes/custom-host,verbs=create;get;update;patch
//+kubebuilder:rbac:groups="",resources=services,verbs=get;list;watch
//+kubebuilder:rbac:groups="",resources=secrets,verbs=get;list;watch;update;patch

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the GitlabPagesSite object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.8.3/pkg/reconcile
func (r *GitlabPagesSiteReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	r.logger = log.Log.WithValues("GitlabPagesSite", req.NamespacedName)
	r.logger.Info("Reconciling GitlabPagesSite")

	gitlabPagesSite := &webservicescernchv1alpha1.GitlabPagesSite{}
	if err := r.Get(context.TODO(), req.NamespacedName, gitlabPagesSite); err != nil {
		if errors.IsNotFound(err) {
			r.logger.Info("GitlabPagesSite resource not found. Ignoring since object must have been deleted.")
			return ctrl.Result{}, nil
		}

		r.logger.Error(err, "Failed to get GitlabPagesSite")
		return ctrl.Result{}, err
	}

	secret, err := r.ensureHttpdSecret(gitlabPagesSite)
	if err != nil {
		// TODO: should we requeue in hope that the Secret will be created in the meantime?
		return ctrl.Result{}, err
	}

	// GitlabPagesSite has been deleted - we should handle the deletion of the entry from the Secret
	if gitlabPagesSite.GetDeletionTimestamp() != nil {
		r.logger.Info("Deleting GitlabPagesSite")

		if controllerutil.ContainsFinalizer(gitlabPagesSite, gitlabPagesSiteFinalizer) {
			// TODO: Remove the entry from the Secret
			delete(secret.Data, generateSecretKey(gitlabPagesSite))

			err := r.Update(context.TODO(), secret, &client.UpdateOptions{})
			if err != nil {
				return ctrl.Result{}, err
			}

			controllerutil.RemoveFinalizer(gitlabPagesSite, gitlabPagesSiteFinalizer)
			err = r.Update(context.TODO(), gitlabPagesSite, &client.UpdateOptions{})
			if err != nil {
				return ctrl.Result{}, err
			}

			route, err := r.getRoute(gitlabPagesSite)
			if err != nil {
				return ctrl.Result{}, err
			}

			err = r.Delete(context.TODO(), route, &client.DeleteOptions{})
			if err != nil {
				return ctrl.Result{}, err
			}
		}

		r.logger.Info("GitlabPagesSite has been deleted")
		return ctrl.Result{}, nil
	}

	appReg, err := r.getApplicationRegistration(ctx, gitlabPagesSite)
	if err != nil {
		return ctrl.Result{}, err
	}

	// Ensure we have a valid ApplicationRegistration registration, this only happens when we have exactly one ApplicationRegistration
	// with a status set by the authz-operator and this status says provisioning succeeded
	if !reflect.DeepEqual(appReg.Status, authzalpha1.ApplicationRegistrationStatus{}) && appReg.Status.ProvisioningStatus != "Created" {
		r.logger.Info(fmt.Sprintf("ApplicationRegistration %v still doesn't have ProvisioningStatus set to Created", appReg.Name))
		return ctrl.Result{RequeueAfter: time.Second * 5}, nil
	}

	oidcSecret := &v1.Secret{}
	namespaceName := types.NamespacedName{
		Namespace: gitlabPagesSite.Namespace,
		Name:      appReg.Status.ClientCredentialsSecret,
	}

	if err := r.Get(ctx, namespaceName, oidcSecret); err != nil {
		r.logger.Error(err, "Failed to the OIDC client secret generated by the ApplicationRegistration CR in the UserProvidedDirectory namespace")
		return ctrl.Result{}, err
	}

	err = r.ensureGitlabPagesSiteHttpdConfig(gitlabPagesSite, secret, oidcSecret)
	if err != nil {
		r.logger.Error(err, "There was a problem while generating httpd config for the GitlabPagesSite")
		return ctrl.Result{}, err
	}

	err = r.ensureRoute(gitlabPagesSite)
	if err != nil {
		r.logger.Error(err, "There was a problem while creating a Route")
		return ctrl.Result{}, err
	}

	if !controllerutil.ContainsFinalizer(gitlabPagesSite, gitlabPagesSiteFinalizer) {
		controllerutil.AddFinalizer(gitlabPagesSite, gitlabPagesSiteFinalizer)

		err = r.Update(context.TODO(), gitlabPagesSite, &client.UpdateOptions{})
		if err != nil {
			return ctrl.Result{}, err
		}
	}

	return ctrl.Result{}, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *GitlabPagesSiteReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&webservicescernchv1alpha1.GitlabPagesSite{}).
		Watches(&source.Kind{Type: &routev1.Route{}}, handler.EnqueueRequestsFromMapFunc(
			func(a client.Object) []reconcile.Request {
				namespace := a.GetNamespace()

				// we only care about the `Route`s within the namespace specified by r.Namespace
				if namespace != r.Namespace {
					return []reconcile.Request{}
				}

				annotations := a.GetAnnotations()
				if annotations != nil && annotations["webservices.cern.ch/static-site-owner"] != "" {
					parts := strings.Split(annotations["webservices.cern.ch/static-site-owner"], "/")
					return []reconcile.Request{{NamespacedName: types.NamespacedName{Name: parts[1], Namespace: parts[0]}}}
				}

				return []reconcile.Request{}
			}),
		).
		Complete(r)
}
