/*


Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package v1alpha1

import (
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

type ResourceCategoryType string

const (
	// The application portal defines these 4 values
	ResourceCategoryUndefined ResourceCategoryType = "Undefined"
	ResourceCategoryTest      ResourceCategoryType = "Test"
	ResourceCategoryPersonal  ResourceCategoryType = "Personal"
	ResourceCategoryOfficial  ResourceCategoryType = "Official"
)

// ApplicationRegistrationSpec defines the desired state of ApplicationRegistration
type ApplicationRegistrationSpec struct {
	// ApplicationName is used to construct the Application's "display name" following a naming convention:
	// `Web frameworks site <applicationName> (<instance>)` (see also https://gitlab.cern.ch/paas-tools/operators/authz-operator/issues/5 ).
	// +kubebuilder:validation:Required
	// +kubebuilder:validation:MinLength=1
	ApplicationName string `json:"applicationName" valid:"matches(^[0-9a-z_\\-]+$),length(1|100)"`
	// Description defines the purpose of the App
	// +kubebuilder:validation:Required
	// +kubebuilder:validation:MinLength=1
	Description string `json:"description"`
	// HomePage points to the app's homepage
	// +optional
	HomePage string `json:"homePage,omitempty"`
	// InitialOwner is the owner defined at creation time; can then be modified at the application portal
	// +kubebuilder:validation:Required
	InitialOwner `json:"initialOwner"`
	// InitialResourceCategory sets the desired resourceCategory on creation of the application registration
	// in the application portal. After creation, the category is managed by the owner in the portal.
	// +kubebuilder:validation:Enum="Undefined";"Test";"Personal";"Official"
	// +kubebuilder:default:="Undefined"
	InitialResourceCategory ResourceCategoryType `json:"initialResourceCategory"`
	// DisplayName is the user-facing name of the ApplicationRegistration as stored in the Authzsvc API
	// +optional
	DisplayName string `json:"displayName,omitempty"`
	// ApplicationIdentifier is the client ID for keycloak. It is auto-generated by the operator following a naming convention,
	// Name convention: `webframeworks-<instance>-<applicationName>` (see also https://gitlab.cern.ch/paas-tools/operators/authz-operator/issues/5 )
	// +optional
	ApplicationIdentifier string `json:"applicationIdentifier,omitempty" valid:"matches(^[a-z][0-9a-z_\\-][0-9a-z_\\-]+$),length(3|127)"`
}

// ApplicationRegistrationStatus defines the observed state of ApplicationRegistration
type ApplicationRegistrationStatus struct {
	// ProvisioningStatus shows whether:
	// - Created: the ApplicationRegistration has been successfully provisioned
	// - Creating: provisioning is still in progress (including any potentially transient errors
	//   from the Authorization API where reconciliation will be reattempted, like
	//   the API not responding or 5xx HTTP errors)
	// - ProvisioningError: provisioning has failed and won't be retried; look at ErrorMessage for details.
	//   It is set for permanent errors, such as name conflicts, invalid initialOwner, 4xx HTTP errors.
	// +kubebuilder:validation:Enum="Created";"Creating";"ProvisioningError";"DeletedFromAPI"
	ProvisioningStatus string `json:"provisioningStatus,omitempty"`

	// ErrorMessage will be set in the event of a terminal error reconciling (or provisioning) the
	// ApplicationRegistration with a human-readable message.
	// This won't be set for transient errors where reconciliation will be reattempted.
	// Error messages originating from the API returned as-is
	// +optional
	ErrorMessage string `json:"errorMessage,omitempty"`
	// ErrorReason will be set in the event of a terminal error reconciling (or provisioning) the
	// ApplicationRegistration with a machine-readable message.
	// This won't be set for transient errors where reconciliation will be reattempted.
	// +optional
	// +kubebuilder:validation:Enum={"ApplicationAlreadyExists","InvalidSpec","k8sAPIClientError", "AuthzAPIClientError","AuthzAPIError","AuthzAPIPermanentError","AuthzAPIInvalidResponse","OwnerNotFound","AuthzsvcApiError","Creating","InvalidOwner","AssociatedApplicationNotFound"}
	ErrorReason string `json:"errorReason,omitempty"`

	// ID is the Unique Identifier of the AplicationRegistration in the Authzsvc API, (This name didn't expect multiple IDs on the Status, therefore the generic ID name)
	// +optional
	ID string `json:"id,omitempty"`
	// CurrentOwnerUsername is the current owner's UPN read from the Application Portal
	// +optional
	CurrentOwnerUsername string `json:"currentOwnerUsername,omitempty"`
	// CurrentAdminGroup is read from the Application Portal
	// +optional
	CurrentAdminGroup string `json:"currentAdminGroup,omitempty"`
	// CurrentResourceCategory is read from the Application Portal
	// +optional
	CurrentResourceCategory ResourceCategoryType `json:"currentResourceCategory,omitempty"`
	// CurrentDescription is read from the Application Portal
	// +optional
	CurrentDescription string `json:"currentDescription,omitempty"`
	// CurrentEnabledStatus shows whether the application is currently enabled
	// +optional
	CurrentEnabledStatus bool `json:"currentEnabledStatus,omitempty"`
	// RegistrationID is the identity of the oidc registration of the application in the Authorization service API
	// +optional
	RegistrationID string `json:"registrationId,omitempty"`
	// OIDCEnabled is a flag to know if the OIDC credentials are enabled
	// +optional
	OIDCEnabled bool `json:"OIDCEnabled,omitempty"`
	// +optional
	TokenExchangePermissions `json:"tokenExchangePermissions,omitempty"`
	// ClientCredentialsSecret is the name of the k8s secret holding the OIDC client credentials
	// +optional
	ClientCredentialsSecret string `json:"clientCredentialsSecret,omitempty"`
	// RedirectURIs is the URI where users will be redirected after authenticating to the IdP during the OIDC flows
	// +optional
	RedirectURIs []string `json:"redirectURIs,omitempty"`
}

// +kubebuilder:object:root=true
// +kubebuilder:subresource:status

// ApplicationRegistration creates and maintains the Application and OIDC registration objects in the AuthzAPI.
// More info: https://gitlab.cern.ch/paas-tools/operators/authz-operator#applicationregistration
type ApplicationRegistration struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   ApplicationRegistrationSpec   `json:"spec,omitempty"`
	Status ApplicationRegistrationStatus `json:"status,omitempty"`
}

// +kubebuilder:object:root=true

// ApplicationRegistrationList contains a list of ApplicationRegistration
type ApplicationRegistrationList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []ApplicationRegistration `json:"items"`
}

func init() {
	SchemeBuilder.Register(&ApplicationRegistration{}, &ApplicationRegistrationList{})
}

// InitialOwner is the owner of the resource at creation time
type InitialOwner struct {
	// username is the owner's CERN username
	// +kubebuilder:validation:Required
	// +kubebuilder:validation:MinLength=1
	Username string `json:"username"`
}

// AppRole is the Application role created by default together with the application.
type AppRole struct {
	Name            string `json:"name"`
	Required        bool   `json:"required"`
	ApplyToAllUsers bool   `json:"applyToAllUsers"`
	Description     string `json:"description"`
	DisplayName     string `json:"displayName"`
	MinLoA          int    `json:"minimumLevelOfAssurance"`
}

type OIDCFlowType string

const (
	OIDCFlowImplicit          OIDCFlowType = "Implicit"
	OIDCFlowAuthorizationCode OIDCFlowType = "AuthorizationCode"
	OIDCFlowClientCredentials OIDCFlowType = "ClientCredentials"
)

// TokenExchangePermissions lists applications for which token exchange permissions have been requested or allowed
type TokenExchangePermissions struct {
	// requests are the Application identifiers for which token exchange permissions are requested
	// +optional
	Requests []string `json:"requests,omitempty"`
	// allowed are the Application identifiers for which token exchange is allowed (has been granted by the corresponding application)
	// +optional
	Allowed []string `json:"allowed,omitempty"`
}

// DisplayNameConvention implements the naming convention {ApplicationName -> DisplayName}
func (a ApplicationRegistration) DisplayNameConvention(clusterInstanceName string) string {
	return "Web frameworks site " + a.Spec.ApplicationName + " (" + clusterInstanceName + ")"
}

// ApplicationIdentifierConvention implements the naming convention {ApplicationName -> ApplicationIdentifier}
func (a ApplicationRegistration) ApplicationIdentifierConvention(clusterInstanceName string) string {
	return "webframeworks-" + clusterInstanceName + "-" + a.Spec.ApplicationName
}
